from .barlowtwins import BarlowTwins
from .byol import BYOL
from .dino import Dino
from .idmm import IDMM
from .simclr import SimCLR
from .simsiam import SimSIAM
from .vicreg import VICReg

__all__ = [
    "BarlowTwins",
    "BYOL",
    "Dino",
    "IDMM",
    "SimCLR",
    "SimSIAM",
    "VICReg",
]
